#include <click/config.h>
#include <click/glue.hh>
#include <click/pair.hh>
#include <click/straccum.hh>
#include <click/string.hh>
#include "bintree.hh"
CLICK_DECLS

void BinContainer::initialize(unsigned int size) {
    data = NULL;
    len = size;
    hash_is_cached = false;
    cached_hash = 0;
}

BinContainer::BinContainer() {
    initialize(0);
}

BinContainer::BinContainer(unsigned int size) {
    initialize(size);
}

BinContainer::~BinContainer() {}

uint8_t *BinContainer::get() const {
    return data;
}

unsigned int BinContainer::getLen() const {
    return len;
}

bool BinContainer::match(uint8_t const* buf, unsigned int size) {
    if (data && size == len) {
	return memcmp(data, buf, len) == 0;
    }
    return false;
}

/* deep probe! */
bool BinContainer::match(BinContainer const& other) {
    if (data == other.get()) {
	if (len == other.getLen())
	    return true;
	else {
	    /* This is weird. Pointer to same memory, but different in size */
	    assert(0);
	    return false;
	}
    } else {
	if (data && len == other.getLen())
	    return memcmp(data, other.get(), len) == 0;
    }
    return false;
}

unsigned int BinContainer::hashvalue() {
    unsigned int res = 0;

    if (hash_is_cached)
	return cached_hash;

    if (len == sizeof(unsigned int)) {
	hash_is_cached = true;
	cached_hash = *((unsigned int*)get());
	return cached_hash;
    }
    
    for(unsigned int i=0;i<len;i++)
	res += *(data+i) * (i+1);

    cached_hash = res;
    hash_is_cached = true;
    return res;
}

BinContainerData::BinContainerData(uint8_t *ptr, unsigned int size) : BinContainer(size) {
    data = new uint8_t[size];
    memcpy(data, ptr, size);
}

BinContainerData::BinContainerData(BinContainer const& other) {
    magic_copy(other);
}

BinContainerData &BinContainerData::operator=(BinContainer const& other) {
    magic_copy(other);
    return *this;
}

BinContainerData::~BinContainerData() {
    if (data != NULL)
	delete [] data;
}

void BinContainerData::magic_copy(BinContainer const& other) {
    if (this != &other) {
	uint8_t *tmp = copy_alloc(other.get(), other.getLen());
	if (tmp == NULL)
	    return;

	if (data != NULL)
	    delete [] data;
	data = tmp;
	len = other.getLen();
	hash_is_cached = false;
    }
}

uint8_t *BinContainerData::copy_alloc(uint8_t *ptr, unsigned int siz) {
    uint8_t *res = new uint8_t[siz];
    memcpy(res, ptr, siz);
    return res;
}

void BinContainerData::set(uint8_t *ptr, unsigned int siz) {
    uint8_t *tmp = copy_alloc(ptr, siz);
    if (tmp == NULL)
	return;
    if (data != NULL)
	delete [] data;
    data = tmp;
    len = siz;
}

void BinContainerData::set(unsigned int val) {
    set(reinterpret_cast<uint8_t*>(&val), sizeof(val));
}



BinContainerRef::BinContainerRef(uint8_t *dat, unsigned int siz) 
  : BinContainer(siz) 
{
    data = dat;
}

BinContainerRef::BinContainerRef(unsigned int *val)
    : BinContainer(sizeof(unsigned int)) 
{
    data = reinterpret_cast<uint8_t*>(val);
}

BinContainerRef::BinContainerRef(BinContainer const& other) {
    if (this != &other) {
	data = other.get();
	len = other.getLen();
    }
}

BinContainerRef::~BinContainerRef() {}

BinContainerRef &BinContainerRef::operator=(BinContainer const& other) {
    if (this != &other) {
	data = other.get();
	len = other.getLen();
    }
    return *this;
}

void BinContainerRef::set(uint8_t *dat, unsigned int siz) {
    data = dat;
    len = siz;
}

void BinContainerRef::set(unsigned int *value) {
    data = reinterpret_cast<uint8_t*>(value);
    len = sizeof(unsigned int);
}


/* TreeNode */

void TreeNode::init() {
    parent = left = right = NULL;
    depth = 0;
}


TreeNode::TreeNode(BinContainer const& k, BinContainer const& v) 
  : key(k), val(v) 
{
    init();
}

TreeNode::~TreeNode() {
    if (left)
	delete left;
    if (right)
	delete right;
    /* key and val are destroyed automagically */
}

void TreeNode::print(StringAccum& sa) {
    sa << "[D: " << depth << " H: " << key.hashvalue() << "] >> (";
    if (left)
	left->print(sa);
    else
	sa << "null";
    
    sa << ",";
    
    if (right)
	right->print(sa);
    else
	sa << "null";
    
    sa << ")";
}

int TreeNode::addNode(TreeNode* node) {
    unsigned int nhash = node->getHashValue();
    TreeNode *ptr = this;
    int nd = 1;

    node->setDepth(0);

    while (1) {
	if (nhash <= ptr->getHashValue()) {
	    if (ptr->left == NULL) {
		node->setParent(ptr);
		ptr->left = node;
		break;
	    }
	    ptr = ptr->left;
	} else {
	    if (ptr->right == NULL) {
		node->setParent(ptr);
		ptr->right = node;
		break;
	    }
	    ptr = ptr->right;
	}
    }

    /* ptr is the parent of the @node */
    do {
	if (nd > ptr->getDepth())
	    ptr->setDepth(nd);
	nd++;
	ptr = ptr->getParent();
    } while(ptr != NULL);

    return getDepth();
}

/* Returns value if key is found in the tree. As a side effect updates depths
   of the tree branches as the search progresses */

BinContainerData* TreeNode::lookup(unsigned int hv, BinContainer const& k) {
    TreeNode *res = this;
    unsigned int chv;
    int cdep = depth;

    while(res != NULL) {
	res->setDepth(cdep++);
	chv = res->getHashValue();
	if (chv == hv) {
	    /* match the whole key */
	    if (key.match(k))
		return &val;
	    /* in case of no match, let's traverse the tree a bit further
	     * to look for duplicates */

	    if (res->getLeft() != NULL) 
		res = res->getLeft();
	    else 
		return NULL;
	    continue;
	}

	if (hv < chv)
	    res = res->getLeft();
	else 
	    res = res->getRight();
    }
    return NULL;
}


/* Returns closest node (first free) for a hash */
TreeNode* TreeNode::lookupNode(unsigned int hash) {
    TreeNode *res = this;
    unsigned int hv;

    while(res != NULL) {
	hv = res->getHashValue();
	if (hv == hash)
	    return res;
	else if (hv > hash) {
	    if (res->getLeft() == NULL)
		return this;
	    else
		res = res->getLeft();
	} else {
	    if (res->getRight() == NULL)
		return this;
	    else
		res = res->getRight();
	}
    }

    return NULL;
}

/*
 * Make my right child my new parent. I'll be connected to the
 * left subtree of my parent.
 * My parent's left subtree will be my new right subtree.
 * M = me, P = parent
 *
 *    M                 P
 *   / \               / \
 *  X   P    ===>     M   Z
 *     / \           / \   \
 *    Y   Z         X   Y   ...
 *         \
 *          ...          
 *
 * precond: 
 * 1. this->getRight() != NULL
 * 
 */
void TreeNode::rightRotate() {
    TreeNode *newp = getRight();
  
    if (getParent()) {
	getParent()->setRight(newp);
    }
    newp->setParent(getParent());

    setParent(newp);

    if (newp->getLeft()) {
	newp->getLeft()->setParent(this);
    }
    setRight(newp->getLeft());
    
    newp->setLeft(this);

    newp->setDepth(newp->getDepth() + 1);
}

/* Move me left subtree above me (degenerate into a list)
 * M = me, P = parent
 *
 *    P                P
 *     \                \
 *      M       ==>      X
 *     / \              / \
 *    X   ...          Y   Z
 *   / \                    \
 *  Y   Z                    M
 *                            \
 *                             ...
 * 
 * precond:
 * 1. this->getLeft() != NULL
 */
void TreeNode::leftRotate() {
    TreeNode *bubbler = this->getLeft();

    bubbler->setParent(this->getParent());
    if (this->getParent() != NULL)
	this->getParent()->setRight(bubbler);

    while(bubbler->getRight() != NULL)
	bubbler = bubbler->getRight();

    bubbler->setRight(this);
    setParent(bubbler);
    setLeft(NULL);
} 

/* BinTree */

void BinTree::balance_tree() {
    TreeNode *ptr = root;
    TreeNode *head;
    int blen = 0;
    int transforms;

    if (entries < 5) {
	click_chatter("Skipping tree balance due to too few entries");
	return;
    }

#ifdef DEBUG_PSCODE
    click_chatter("Balancing tree");
#endif

    /* degenerate tree bottom-up */
    while(ptr->getRight() != NULL)
	ptr = ptr->getRight();

    while(1) {
	ptr->setDepth(0);
	if (ptr->getLeft()) {
	    ptr->leftRotate();
	}

	if (ptr->getParent() == NULL)
	    break;
	ptr = ptr->getParent();
    }	

    head = ptr;

    blen = entries;
    int do_until = simple_log2(entries);

    for(; blen > do_until; blen -= (blen-1)/2) {
	TreeNode *tmp = head;
	transforms = (blen-1)/2;
	while (transforms > 0) {
	    if (!tmp || !tmp->getRight()) {
		break;
	    }

	    tmp->rightRotate();
	    tmp = tmp->getParent()->getRight();
	    transforms--;
	}


	while(head->getParent() != NULL)
	    head = head->getParent();

    }

    root = head;

    /* Assumes there is a right child, and there should always be */
    root->setDepth(root->getRight()->getDepth() + 1);

    /* the tree should be quite balanced now.
     * The depths are updated only on the nodes that are new parents in
     * right rotations. However, the new root will be one of these, so
     * the other do not matter.
     */
#ifdef DEBUG_PSCODE
    click_chatter("Tree balanced");
    {
	StringAccum sa;
	root->print(sa);
	click_chatter("Tree: %s", sa.take_string().c_str());
    }

    if (!consistency_check(root)) {
	click_chatter("Failed tree consistency check after balancing operation");
	_outoforder = true;
    }
#endif
}

#ifdef DEBUG_PSCODE
void BinTree::print_relatives(TreeNode *parent, TreeNode *left, TreeNode *right) {
    if (parent)
	click_chatter("==> Parent: %d", parent->getHashValue());
    else
	click_chatter("==> Parent: None");
    if (left)
	click_chatter("==> Left  : %d", left->getHashValue());
    else
	click_chatter("==> Left  : None");
    if (right)
	click_chatter("==> Right : %d", right->getHashValue());
    else
	click_chatter("==> Right : None");
}

bool BinTree::consistency_check(TreeNode *nod) {
    TreeNode *parent;
    TreeNode *left;
    TreeNode *right;

    if (nod == NULL) {
	click_chatter("Node is NULL assertation:");
	return false;
    }

    parent = nod->getParent();
    left = nod->getLeft();
    right = nod->getRight();

    click_chatter("== Current Node: %d", nod->getHashValue());

    /* test binary tree invariants */
    if (left) {
	if (left->getParent() != nod) {
	    click_chatter("I'm not parent of my left child!");
	    print_relatives(parent, left, right);
	    click_chatter("Left child's view: (%d)", left->getHashValue());
	    print_relatives(left->getParent(), left->getLeft(), left->getRight());
	    return false;
	}

	if (left->getHashValue() > nod->getHashValue()) {
	    click_chatter("BT Invariant failed for left child");
	    print_relatives(parent, left, right);
	    return false;
	}
    }

    if (right) {
	if (right->getParent() != nod) {
	    click_chatter("I'm not parent of my right child!");
	    print_relatives(parent, left, right);
	    click_chatter("Right child's view: (%d)", right->getHashValue());
	    print_relatives(right->getParent(), right->getLeft(), right->getRight());
	    return false;
	}

	if (right->getHashValue() <= nod->getHashValue()) {
	    click_chatter("BT Invariant failed for right child");
	    print_relatives(parent, left, right);
	    return false;
	}
    }

    bool res = true;

    /* unfortunately recursive */
    if (left)
	res = consistency_check(left);
    if (right)
	res = (res && consistency_check(right));
    
    return res;
}
#endif

int BinTree::log_helper(unsigned int val) {
    struct {
	int mask;
	int val;
    } t[] = { 
	{0x80, 7}, {0x40, 6}, {0x20, 5}, {0x10, 4}, 
	{0x08, 3}, {0x04, 2}, {0x02, 1}, {0x01, 0} 
    };
    
    for(int i=0;i<8;i++) {
	if (val & t[i].mask)
	    return t[i].val;
    }
    return 0;
}

/* = floor(log2(X)) */
int BinTree::simple_log2(unsigned int val) {
    if (val < 0xFF)
	return log_helper(val);
    if (val < 0xFFFF)
	return log_helper((val >> 8)) + 8;
    if (val < 0xFFFFFF)
	return log_helper((val >> 16)) + 16;
    return log_helper((val >> 24)) + 24;
}
    
BinTree::BinTree(int limit) {
    elimit = limit;
    root = NULL;
    entries = 0;
    _outoforder = false;
}


BinTree::~BinTree() {
    delete root;
}

void BinTree::addEntry(BinContainer const& key, BinContainer const& value) {
    if (_outoforder)
	return;

    TreeNode* node = NULL;
    if (entries >= elimit) {
	/* just replace one... but replace it intelligently:
	 * if possible, replace so that the tree structure doesn't require modification 
	 */
	unsigned int hv = const_cast<BinContainer&>(key).hashvalue();
	node = root->lookupNode(hv);
	if (const_cast<BinContainerData&>(node->getKey()).hashvalue() == hv) {
	    if (node->isSameKey(key)) {
		return;
	    } else {
		node->replaceValue(value);
		return;
	    }
	}
	/* replace both */
#ifdef DEBUG_PSCODE
	click_chatter("Replacing: %d with %d", node->getHashValue(), 
		      const_cast<BinContainer&>(key).hashvalue());
#endif
	node->replaceKey(key);
	node->replaceValue(value);
	return;
    }
    
    node = new TreeNode(key, value);
    if (entries++ == 0)
	root = node;
    else
	root->addNode(node);
        
    if (root->getDepth() > simple_log2(entries)+1 || entries >= elimit)
	balance_tree();
    
}

BinContainerData* BinTree::lookupEntry(BinContainer const& key) { 
    if (_outoforder)
	return NULL;

    if (root == NULL)
	return NULL;
    int v = const_cast<BinContainer &>(key).hashvalue();
    return root->lookup(v, key);
}

String BinTree::print() {
    StringAccum sa;
    sa << "Bintree contains: " << entries << " elements (out of maximum " << elimit << " )" << "\n";
    if (_outoforder) {
	sa << "Bintree is out of order\n";
	return sa.take_string();
    }

    if (root != NULL)
	root->print(sa);
    return sa.take_string();
}

CLICK_ENDDECLS
ELEMENT_PROVIDES(BinTree)
