#ifndef _BINTREE_HH_
#define _BINTREE_HH_

#include <click/glue.hh>
#include <click/pair.hh>
#include <click/straccum.hh>
#include <click/string.hh>

CLICK_DECLS

typedef unsigned char uint8_t;

/**
 * BinContainer is a pure virtual class, which has two subclasses:
 * BinContainerData and BinContainerRef. The former copies the data,
 * while the latter just refers to the user supplied data.
 * 
 * Usually lookup operations are done using BCRefs and stores 
 * using BCDatas.
 *
 */
class BinContainer {
    uint8_t *data;
    unsigned int len;
    unsigned int cached_hash;
    bool hash_is_cached;

private:
    void initialize(unsigned int size);
public:
    BinContainer();
    explicit BinContainer(unsigned int size);
    virtual ~BinContainer();

    virtual bool isReference() const  = 0;
    uint8_t *get() const;
    unsigned int getLen() const;
  /* Deep match = if key's hashvalue matches, match also the whole key */
    bool match(BinContainer const& other);
    bool match(uint8_t const* od, unsigned int ol);

    unsigned int hashvalue();
    friend class BinContainerData;
    friend class BinContainerRef;
};

class BinContainerData : public BinContainer {
private:
    void magic_copy(BinContainer const& other);
    uint8_t *copy_alloc(uint8_t *p, unsigned int l);
public:
    BinContainerData(uint8_t *buf, unsigned int size);
    BinContainerData(BinContainer const& other);
    ~BinContainerData();

    BinContainerData& operator=(BinContainer const &other);
    void set(uint8_t * nd, unsigned int nl);
    void set(unsigned int value);
    bool isReference() const { return false; }
};

class BinContainerRef : public BinContainer {
public:
    BinContainerRef(unsigned int *val);
    BinContainerRef(uint8_t *buf, unsigned int size);
    BinContainerRef(BinContainer const& other);
    ~BinContainerRef();

    BinContainerRef& operator=(BinContainer const &other);
    void set(uint8_t * nd, unsigned int nl);
    void set(unsigned int *value);
    bool isReference() const { return true; }
};



class TreeNode {
    TreeNode* left;
    TreeNode* right;
    TreeNode* parent;
    BinContainerData key;
    BinContainerData val;

    void init();
    
  /* =max(depth(left), depth(right) */
    int depth;
private:
    void print(StringAccum& sa);
    /* Find a node whose hashvalue is the closest match to the argument */
    TreeNode* lookupNode(unsigned int hash);
    /* See bintree.cc for *Rotate() */
    void rightRotate();
    void leftRotate();
public:

    TreeNode(BinContainer const& k, BinContainer const& v);
    ~TreeNode();

    int addNode(TreeNode* node);
    BinContainerData* lookup(unsigned int hv, BinContainer const& k);

    inline bool isSameKey(BinContainer const& other) { return key.match(other); }
    inline void replaceKey(BinContainer const& with) { key = with; }
    inline void replaceValue(BinContainer const& with) { val = with; }

    inline unsigned int getHashValue() { return key.hashvalue(); }

    inline TreeNode* getLeft() { return left; }
    inline TreeNode* getRight() { return right; }
    inline TreeNode* getParent() { return parent; }

    inline void setLeft(TreeNode *n) { left = n; }
    inline void setRight(TreeNode *n) { right = n; }
    inline void setParent(TreeNode *n) { parent = n; }

    inline void setDepth(int n) { depth = n; }
    inline int getDepth() { return depth; }

    inline BinContainerData const& getKey() { return key; }
    inline BinContainerData const& getVal() { return val; }

    friend class BinTree;
};

class BinTree {
    int entries;
    int elimit;
    TreeNode* root;
    bool _outoforder;

    int log_helper(unsigned int val);
    /* =floor(log2(val)) */
    int simple_log2(unsigned int val);
    /* Run tree balancing operation [O(n) time, O(1) space] */
    void balance_tree();
#ifdef DEBUG_BINTREE
    void print_relatives(TreeNode *, TreeNode *, TreeNode *);
    bool consistency_check(TreeNode *);
#endif
public:
    BinTree(int limit);
    ~BinTree();

    void addEntry(BinContainer const& key, BinContainer const& value);
    BinContainerData* lookupEntry(BinContainer const& key);
    String print();
};

CLICK_ENDDECLS
#endif /* _BIN_TREE_H_ */

