#!/usr/bin/env python2.6

#
# Copyright (C) 2009-2010, Oy L M Ericsson Ab, NomadicLab.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version
# 2 as published by the Free Software Foundation.
#
# Alternatively, this software may be distributed under the terms of
# the BSD license.
#
# See LICENSE and COPYING for more details.
#

# Packet generator utility
# ------------------------
#
# Do note that the FreeBSD version requires that the libpsirp API for
# Python is compiled with socket support enabled and that the psfs
# kernel module is loaded.


import socket, struct, sys, time, traceback
import binascii


PLATFORM_LINUX   = 0
PLATFORM_FREEBSD = 1

def detect_platform():
    if sys.platform.startswith("linux2"):
        return PLATFORM_LINUX
    elif sys.platform.startswith("freebsd"):
        return PLATFORM_FREEBSD
    else:
        assert False, "Platform not supported: %r" % sys.platform
platform = detect_platform()

if platform == PLATFORM_FREEBSD:
    from psirp.libpsirp_py import *


ETH_PS_PROTO = 0xACDC # our ethertype
ETH_DATA_LEN = 1500  # max eth data size
ETH_HDR_LEN = 14 # eth header size
ETH_BC_ADDR = 6*"\xff" # eth broadcast
#AF_LINK = 18
FWDHDR_SIZE = 42
RZVHDR_SIZE = 112
METAHDR_SIZE = 152
hw_addr=""
raw_socket = None
verbose = False

global_fid  = binascii.unhexlify("00aa000000000000000000000000000000000000000000000000000000000000")
global_sid  = binascii.unhexlify("0000000000000000000000000000000000000000000000000000000000000001")
global_rid  = binascii.unhexlify("0000000000000000000000000000000000000000000000000000000000000555")
global_vrid = binascii.unhexlify("0000000000000000000000000000000000000000000000000000000000001000")

def add_fwdhdr(next, fid, frame, offset):
    struct.pack_into("!BBBBB5s32s", frame, offset,
                     0x01, next, 255, 0, 1, 5*b'\x00', fid)

def add_rzvhdr(subscribe, next, sid, rid, vrid, seqnum, frame, offset):
    if subscribe == True:
        mode = 0x02
    else:
        mode = 0x04
    struct.pack_into("!BB32s32s32sQB5s", frame, offset,
                     mode, next, sid, rid, vrid, seqnum, 1, 5*b'\x00')

def add_metahdr(next, vrid, size, maxseq, frame, offset):
    struct.pack_into("!BB32s32s32s32sQQBB4s", frame, offset,
                     0x06, next, global_sid, global_rid, vrid, global_fid,
                     1024, 1, 0, 0, 4*b'\x00')
    
def create_publish_data_frame(vrid, seqnum, ssize):
    newframe = bytearray((ETH_HDR_LEN + FWDHDR_SIZE + RZVHDR_SIZE + ssize))
    
    struct.pack_into("!6s6sH", newframe, 0,
                     ETH_BC_ADDR, hw_addr, ETH_PS_PROTO)

    add_fwdhdr(0x04, global_fid, newframe, ETH_HDR_LEN)
    add_rzvhdr(False, 0xF0, global_sid, global_rid, vrid, seqnum, newframe,
               ETH_HDR_LEN+FWDHDR_SIZE)

    return newframe

def create_subscribe_data_frame(vrid, size, seqnum, maxseq, filling=0):
    newframe = bytearray(ETH_HDR_LEN + FWDHDR_SIZE + RZVHDR_SIZE + METAHDR_SIZE)

    struct.pack_into("!6s6sH", newframe, 0,
                     ETH_BC_ADDR, hw_addr, ETH_PS_PROTO)
    add_fwdhdr(0x02, global_fid, newframe, ETH_HDR_LEN)
    add_rzvhdr(True, 0x06, global_sid, global_rid, vrid, seqnum, newframe,
               ETH_HDR_LEN+FWDHDR_SIZE)
    add_metahdr(0xF0, vrid, size, maxseq,
                newframe, ETH_HDR_LEN+FWDHDR_SIZE+RZVHDR_SIZE)

    return newframe
  
    
def send_frame(frame):
    global verbose
    try:
        if platform == PLATFORM_FREEBSD:
            psirp_py_sock_send(sockdata, frame)
        elif platform == PLATFORM_LINUX:
            raw_socket.send(frame)
            
        if verbose:
            print "sent frame"
    except Exception, e:
        traceback.print_exc()


# id = (H)*(::){0-1}(H)*
# H = [0-9A-F]
# 2 <= len(id) <= 64
def parse_id(mid):
    global global_rid

    if len(mid) < 2 or len(mid) > 64:
        assert False, "id of wrong dimensions"

    colon = mid.find("::")
    if colon != -1:
        newstr = mid[:colon]
        rest = len(mid) - colon - 2
        newstr += (64-rest-colon)*'0'
        newstr += mid[colon+2:]
    else:
        newstr = mid

    if len(newstr) != 64:
        print "Error: id string: %s, len=%d" % (newstr, len(newstr))
        assert False, ""
    
    res = binascii.unhexlify(newstr)
    if len(res) != 32:
        print "unhexlified version wrong: ", len(res)
        print "hexlified string: ", newstr
        assert False, ""

    return res


def carry(val, index):
    if  index > 31:
        print "DEBUG: VRID overflow (no error)"
        return

    try:
        val[index] += 1
    except ValueError, e:
        carry(val, index+1)
        val[index] = 0

def calc_vrid(buf):
    val = bytearray(32)
    for i in buf:
        try:
            val[0] += ord(i)
        except ValueError, e:
            carry(val, 1)
            old = val[0]
            val[0] = ord(i) - (0xFF - old + 1)

    print "DEBUG: vrid=%s" % binascii.hexlify(buffer(val))
    return str(val)

def send_frames(if_name, verbose=False, dfile=None, vrid=global_vrid, get=False, size=1024):
    global raw_socket
    if platform == PLATFORM_FREEBSD:
        frame_len = ETH_DATA_LEN
    elif platform == PLATFORM_LINUX:
        raw_socket = socket.socket(socket.AF_PACKET, socket.SOCK_RAW,
                                   ETH_PS_PROTO)
        frame_len = ETH_HDR_LEN + ETH_DATA_LEN
    
    try:
        if platform == PLATFORM_FREEBSD:
            sockdata = psirp_py_sock_create(if_name)
            data_offset = 0
        elif platform == PLATFORM_LINUX:
            raw_socket.bind((if_name, ETH_PS_PROTO))
            sockaddr = raw_socket.getsockname()
            hw_addr = sockaddr[4]
                  
        csize = size
        seq = 0
        if get == True:
            while csize > 0:
                frame = create_subscribe_data_frame(vrid, size,
                                                    seq, size/1024)

                send_frame(frame)
                seq = seq + 1
                csize = csize - 1024
                print "Sent subscribe data frame"
        else:
            if dfile != None:
                dfile.seek(0,2)
                filesize = dfile.tell()
                dfile.seek(0,0)
                if size > filesize:
                    size = filesize
                csize = size

            while csize > 0:
                if csize < 1024:
                    ssize = csize
                else:
                    ssize = 1024
                if dfile != None:
                    buf = dfile.read(ssize)
                else:
                    buf = ssize*b'\x01'
                    
                vrid = calc_vrid(buf)
                frame = create_publish_data_frame(vrid, seq, ssize)
                frame[ETH_HDR_LEN+FWDHDR_SIZE+RZVHDR_SIZE:] = buf
                send_frame(frame)
                seq = seq + 1
                csize = csize - ssize
                print "Published data"


    finally:
        if platform == PLATFORM_FREEBSD:
            psirp_py_sock_close(sockdata)
        if platform == PLATFORM_LINUX:
            raw_socket.close()

    print "Done."

def main():
    import getopt
    
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], "i:f:r:s:hvg", ["help"])
        #print opts, args
    except getopt.GetoptError, err:
        print str(err)
        sys.exit(2)
    
    if_name, kwargs = None, {}
    for o, a in opts:
        if o in ("-h", "--help"):
            print str(__file__), "-i interface -f input file -g subscribe -r rid"
            print str(__file__), "-s max size"
            sys.exit()
        elif o == "-i":
            if_name = a
        elif o == "-v":
            kwargs["verbose"] = True
        elif o == "-f":
            kwargs["dfile"] = file(a, "r")
        elif o == "-g":
            kwargs["get"] = True
        elif o == "-s":
            kwargs["size"] = int(a)
        elif o == "-r":
            kwargs["vrid"] = parse_id(a)
        else:
            assert False, "Unhandled option: %r" % o
    
    assert if_name is not None, "interface name is mandatory (-i interface)"
    
    send_frames(if_name, **kwargs)
        
if __name__ == "__main__":
    main()
