#include <click/config.h>
#include <click/confparse.hh>
#include "pscheck.hh"
#include "pscache.hh"
#include "bintree.hh"

CLICK_DECLS

/* Packetcache
 *
 * Configuration: 
 *
 *
 */

PSCache::PSCache()
{
    /* Our cache has space for 25 entries */
    _tree = new BinTree(25);
}

PSCache::~PSCache()
{
    delete _tree;
}

int PSCache::configure(Vector<String> &, ErrorHandler *)
{
    return 0;
}

void PSCache::push(int, Packet *p) 
{
    click_chatter("Packets should not be pushed to PSCache element!");
    p->kill();
    return;
}

void PSCache::store(unsigned int hashvalue, unsigned char const* start, unsigned int len)
{
    BinContainerRef value(const_cast<uint8_t*>(start), len);
    BinContainerRef key(&hashvalue);

    _tree->addEntry(key, value);
    /* bintree will make copies out of them */
}

String PSCache::print()
{
    if (_tree) {
	String s = String(_tree->print());
	return s;
    }
    return String("N/A");
}

Packet *PSCache::load(unsigned int hashvalue, Packet *old)
{
    unsigned char type = old->anno_u8(PS_ANNO_PKTTYPE);
    WritablePacket *wp;
    BinContainerRef key(&hashvalue);
    const int hlen = sizeof(struct pubsub_fwdhdr) + sizeof(struct pubsub_rzvhdr);

    if ((type & 0x82) != 0x82) {
	String debug("Only type \"SUBSCRIBE DATA\" is supported. Received data: ");
	click_chatter(PSCheck::make_type_string(debug, old));
	return NULL;
    }

    BinContainerData* val = _tree->lookupEntry(key);

    if (val == NULL) {
	click_chatter("Cache miss!");
	return NULL;
    }

    click_chatter("Cache hit!");
	
    wp = WritablePacket::make(val->getLen() + hlen);
    if (!wp) {
	click_chatter("Unable to create packet with cached content [memory?]");
	return NULL;
    }

    /* let's first declare all the headers, otherwise this will look quite messy */
    struct pubsub_fwdhdr *nfwd; 
    const struct pubsub_fwdhdr *ofwd;
    struct pubsub_rzvhdr *nrzv;
    const struct pubsub_rzvhdr *orzv;
    const struct pubsub_metahdr *ometa;

    /* MAC header must be pushed (into headroom) after cache has created the packet */

    ofwd = reinterpret_cast<const struct pubsub_fwdhdr *>(old->mac_header()+14);
    nfwd = reinterpret_cast<struct pubsub_fwdhdr *>(wp->data());
    /* XXX: assuming headers come in this order! */
    orzv = reinterpret_cast<const struct pubsub_rzvhdr *>(ofwd+1);
    ometa = reinterpret_cast<const struct pubsub_metahdr *>(orzv+1);

    nfwd->hdr_type = PSIRP_HDR_FWD_BF;
    nfwd->nxt_hdr = PSIRP_HDR_RZV_PUBLISH_DATACHUNK;
    nfwd->ttl = 64;
    nfwd->d = ofwd->d;               /* reasonable ? */
    nfwd->proto_ver = 1;
    memcpy(nfwd->fid, ometa->fid, sizeof(pubsub_id_t));
    memset(nfwd->reserved, 0, sizeof(nfwd->reserved));

    nrzv = reinterpret_cast<struct pubsub_rzvhdr *>(nfwd+1);

    nrzv->hdr_type = PSIRP_HDR_RZV_PUBLISH_DATACHUNK;
    nrzv->nxt_hdr = PSIRP_PAYLOAD;
    memcpy(nrzv->sid, orzv->sid, sizeof(pubsub_id_t));
    memcpy(nrzv->rid, orzv->rid, sizeof(pubsub_id_t));
    memcpy(nrzv->vrid, orzv->vrid, sizeof(pubsub_id_t));
    nrzv->seqnum = orzv->seqnum;
    nrzv->proto_ver = 0x01;
    memset(nrzv->reserved, 0, sizeof(nrzv->reserved));

    click_chatter("Loading packet from cache:");
    click_chatter("wp->data=%p, nrzv+1=%p, val->data=%p, val->size=%d", wp->data(), nrzv+1, val->get(), val->getLen());
    memcpy(reinterpret_cast<unsigned char *>(nrzv+1), val->get(), val->getLen());

    /* XXX: Rethink packet pointers! */
    return wp;  
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(BinTree PSCheck)
EXPORT_ELEMENT(PSCache)
