#include <click/config.h>
#include <click/confparse.hh>
#include <clicknet/ether.h>
#include "psclassify.hh"

CLICK_DECLS

PSCheck::PSCheck()
{
}

PSCheck::~PSCheck()
{
}

int PSCheck::configure(Vector<String> &, ErrorHandler *)
{
    return 0;
}

bool PSCheck::check_rzv_hdr(const struct pubsub_rzvhdr *rzvp, Packet *p)
{
    /* consistency */
    if (rzvp->nxt_hdr != PSIRP_PAYLOAD &&
	rzvp->nxt_hdr != (rzvp+1)->hdr_type) {
	click_chatter("Malformed rendezvous header");
	return false;
    }

    if (rzvp->proto_ver != PSIRP_RZV_PROTO) {
	click_chatter("Wrong rendezvous protocol");
	return false;
    }

    switch(rzvp->hdr_type) {
    case PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK:
	SET_PS_ANNO_PKTTYPE(p, DATA, SUBSCRIBE);
	break;
    case PSIRP_HDR_RZV_SUBSCRIBE_METADATA:
	SET_PS_ANNO_PKTTYPE(p, META, SUBSCRIBE);
	break;
    case PSIRP_HDR_RZV_PUBLISH_DATACHUNK:
	SET_PS_ANNO_PKTTYPE(p, DATA, PUBLISH);
	break;
    case PSIRP_HDR_RZV_PUBLISH_METADATA:
	SET_PS_ANNO_PKTTYPE(p, META, PUBLISH);
    default:
	break;
    }

    return true;
}

bool PSCheck::check_meta_hdr(const struct pubsub_metahdr *, Packet *)
{
    return true;
}

void PSCheck::push(int, Packet *p) 
{
    const struct click_ether *ethp = reinterpret_cast<const struct click_ether *>(p->mac_header());

    const unsigned char *datap;
    unsigned char nxt = 0;

    if (ethp->ether_type != PUBSUB_PROTOCOL) {
	output(1).push(p);
	return;
    }

    p->pull(sizeof(struct click_ether));
    int offset = p->data() - reinterpret_cast<const unsigned char*>(ethp);

    p->set_anno_u16(PS_ANNO_FWDHDR, 0);
    p->set_anno_u16(PS_ANNO_RZVHDR, 0);
    p->set_anno_u16(PS_ANNO_METAHDR, 0);
    p->set_anno_u8(PS_ANNO_PKTTYPE, 0);

    if (offset != sizeof(struct click_ether)) 
    {
	click_chatter("Pubsub header does not start immediately after ethernet "
		      "header!");
	output(1).push(p);
	return;
    }

    if (p->length() < sizeof(struct pubsub_fwdhdr)) {
	click_chatter("Packet too short");
	output(1).push(p);
	return;
    }

    const struct pubsub_fwdhdr *fwd;
    fwd = reinterpret_cast<const struct pubsub_fwdhdr *>(p->data());
    if (fwd->hdr_type != PSIRP_HDR_FWD_BF) {
	click_chatter("FWD header not first in pubsub packet!");
	output(1).push(p);
	return;
    }

    p->set_anno_u16(PS_ANNO_FWDHDR, offset);
    
    if (fwd->proto_ver != PSIRP_FWD_PROTO_BF) {
	click_chatter("Wrong protocol version in forwarding header");
	output(1).push(p);
	return;
    }

    /* consistency first */
    if (fwd->nxt_hdr != (fwd+1)->hdr_type) {
	click_chatter("Malformed pubsub packet");
	output(1).push(p);
	return;
    }

    nxt = fwd->nxt_hdr;
    p->pull(sizeof(struct pubsub_fwdhdr));
    datap = p->data();

    /* check for rzv header */
    if (*datap > 0x01 && *datap < 0x06) {
	if (nxt <= 0x01 && nxt >= 0x06) {
	    click_chatter("Next header mismatch (doesn't indicate RZV header)");
	    output(1).push(p);
	    return;
	}

	if (p->length() < sizeof(struct pubsub_rzvhdr)) {
	    click_chatter("Too short packet (rzvhdr)");
	    output(1).push(p);
	    return;
	}

	p->set_anno_u16(PS_ANNO_RZVHDR, p->data() - p->mac_header());
	if (check_rzv_hdr(reinterpret_cast<const struct pubsub_rzvhdr*>(datap),
			  p) == false) {
	    output(1).push(p);
	    return;
	}

	nxt = *(datap+1);
	p->pull(sizeof(pubsub_rzvhdr));
	datap = p->data();
    }

    /* check for metadata header */
    if (*datap == 0x06 && nxt == 0x06) {
	if (p->length() < sizeof(struct pubsub_metahdr)) {
	    click_chatter("Too short packet (metahdr)");
	    output(1).push(p);
	    return;
	}

	p->set_anno_u16(PS_ANNO_METAHDR, p->data() - p->mac_header());
	if (check_meta_hdr(reinterpret_cast<const struct pubsub_metahdr*>(datap), p) == false) {
	    output(1).push(p);
	    return;
	}
	nxt = *(datap+1);
	p->pull(sizeof(struct pubsub_metahdr));
	datap = p->data();
    }

    if (nxt != 0xF0) {
	click_chatter("Payload not found: %02x", nxt);
	output(1).push(p);
	return;
    }

    output(0).push(p);
    return;
}

#if 1
const char *PSCheck::make_type_string(String& buffer, Packet *p)
{
    unsigned char pkttype = p->anno_u8(PS_ANNO_PKTTYPE);

    if ((pkttype & 0x80) == 0) {
	buffer += "UNCLASSIFIED";
    } else {
	if ((pkttype & 0x01) == 1)
	    buffer += "SUBSCRIBE ";
	else
	    buffer += "PUBSLISH ";

	if ((pkttype & 0x02) == 0)
	    buffer += "META ";
	else
	    buffer += "DATA ";
    }

    return buffer.c_str();
}
#endif

CLICK_ENDDECLS
EXPORT_ELEMENT(PSCheck)
