#ifndef CLICK_PSCHECK_HH
#define CLICK_PSCHECK_HH
#include <click/element.hh>
#include <click/atomic.hh>
CLICK_DECLS

/*
=c

PSClassify()

=s pubsub

Calculates hash suitable for PSCache element. Sets the result in packet's annotations.

=d

Hash is calculated from the OFFSET of the beginning of the packet, and SIZE consequtive
bytes are used.

The hash result value is added to packet's annotations, otherwise packet is not modified.

*/
#define PUBSUB_PROTOCOL (htons(0xACDC))

/* Forwarding header */
#define PSIRP_HDR_FWD_BF                  0x01
/* Rendezvouys headers */
#define PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK 0x02
#define PSIRP_HDR_RZV_SUBSCRIBE_METADATA  0x03
#define PSIRP_HDR_RZV_PUBLISH_DATACHUNK   0x04
#define PSIRP_HDR_RZV_PUBLISH_METADATA    0x05
#define PSIRP_HDR_RZV_NOTIFICATION        0x15
#define PSIRP_HDR_RZV_PUBLISH_BROADCAST   0x17
#define PSIRP_HDR_RZV_PUBLISH_LSA         0x18
/* Metadata header */
#define PSIRP_HDR_MD                      0x06
#define PSIRP_HDR_RZVSIGNAL               0x16
/* Security header */
#define PSIRP_HDR_SEC_PLA                 0x07
/* Payload */
#define PSIRP_PAYLOAD                     0xF0

#define PSIRP_FWD_PROTO_BF            0x01
#define PSIRP_RZV_PROTO               0x01

typedef unsigned char pubsub_id_t[32];

struct pubsub_fwdhdr {
    unsigned char hdr_type;
    unsigned char nxt_hdr;
    unsigned char ttl;
    unsigned char d;
    unsigned char proto_ver;
    unsigned char reserved[5];
    pubsub_id_t   fid;
} CLICK_SIZE_PACKED_ATTRIBUTE;
/* 42 */

struct pubsub_rzvhdr {
    unsigned char hdr_type;
    unsigned char nxt_hdr;
    pubsub_id_t   sid;
    pubsub_id_t   rid;
    pubsub_id_t   vrid;
    unsigned long long seqnum;
    unsigned char proto_ver;
    unsigned char reserved[5];
} CLICK_SIZE_PACKED_ATTRIBUTE;
/* 112 */

struct pubsub_metahdr {
    unsigned char hdr_type;
    unsigned char nxt_hdr;
    pubsub_id_t   sid;
    pubsub_id_t   rid;
    pubsub_id_t   vrid;
    pubsub_id_t   fid;
    unsigned long long len;
    unsigned long long max_seqnum;
    unsigned char rzvhdr_type;
    unsigned char flags;
    unsigned char reserved[4];
} CLICK_SIZE_PACKED_ATTRIBUTE;
/* 150 */

#define PS_ANNO_FWDHDR 16
#define PS_ANNO_RZVHDR 18
#define PS_ANNO_METAHDR 20

/* 8 bits: 100000TM
 * T = type (meta = 0, data = 1)
 * M = mode (publish = 0, subscribe = 1);
 */
#define PS_ANNO_PKTTYPE 22
#define PS_ANNO_HASHVALUE 23
#define MAKE_PS_ANNO_PKTTYPE(type, mode) (1 << 7 | type << 1 | mode)
#define SET_PS_ANNO_PKTTYPE(packet, type, mode) (packet->set_anno_u8(PS_ANNO_PKTTYPE, MAKE_PS_ANNO_PKTTYPE(type, mode)))

class PSCheck : public Element { 
public:

    enum { PUBLISH, SUBSCRIBE };
    enum { META, DATA };

    PSCheck();
    ~PSCheck();
    
    const char *class_name() const		{ return "PSCheck"; }
    const char *port_count() const		{ return "1/2"; }
    const char *processing() const		{ return "h/h"; }
    const char *flow_code() const               { return "x/xy"; }

    int configure(Vector<String> &, ErrorHandler *);
    void push(int, Packet *);

    static const char *make_type_string(String &, Packet *);
private:
    bool check_rzv_hdr(const struct pubsub_rzvhdr *, Packet *p);
    bool check_meta_hdr(const struct pubsub_metahdr *, Packet *p);
};

CLICK_ENDDECLS
#endif
