#include <click/config.h>
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/string.hh>
#include <click/straccum.hh>
#include "pscheck.hh"
#include "psclassify.hh"

CLICK_DECLS

static Vector<SimpleProg::tokens *> *progwords; 

/* Packetcache
 *
 * Configuration: 
 *
 *
 */

PSClassify::PSClassify() 
{
    sips = new Vector<SimpleProg *>();
    progwords = new Vector<SimpleProg::tokens *>();
    progwords->push_back(new SimpleProg::tokens("PUB", 1));
    progwords->push_back(new SimpleProg::tokens("SUB", 2));
    progwords->push_back(new SimpleProg::tokens("META",3));
    progwords->push_back(new SimpleProg::tokens("DATA",4));
    progwords->push_back(new SimpleProg::tokens("&&",5));
    progwords->push_back(new SimpleProg::tokens("||",6));
    progwords->push_back(new SimpleProg::tokens("!",7));
    progwords->push_back(new SimpleProg::tokens("END",128));
}

PSClassify::~PSClassify()
{
    delete sips;
    delete progwords;
}

int 
PSClassify::create_program(Vector<String> &conf, ErrorHandler *)
{
    SimpleProg *sp;
    if (conf.size() < 1)
	return 0;
    
    int a, b;
    a=b=0;
    bool neg = false;
    for(int i=0;i<conf.size();i++) {
	String word = conf[i];
	sp = new SimpleProg();
	
	while(word.length() > 1) {
	    bool found = false;
	    for(int j=0;j<progwords->size();j++) {
		SimpleProg::tokens *tk = (*progwords)[j];
		if (word.starts_with(tk->token) == true) {
    
		    if (neg && (tk->value < 1 || tk->value > 4)) {
			click_chatter("Invalid program: (PUB, SUB, DATA, META) must follow negation");
			found = false;
			break;
		    }
		    if (tk->value == 7)
			neg = true;
		    else
			neg = false;

		    sp->add_cmd(tk->value, i);
		    a = tk->token.length();
		    /* skip space */
		    while(word.length() > a && word.at(a) == ' ') a++;

		    word = word.substring(a, word.length()-a);
		    found = true;
		    break;
		}
	    }
	    if (found == false) {
		click_chatter("Invalid program keyword: %s", word.c_str());
		delete sp;
		return -1;
	    }
	}
	if (!sp->terminate_cmd(i)) {
	    delete sp;
	    return -1;
	}
	sips->push_back(sp);
    }
    
    return sips->size();
}


int PSClassify::configure(Vector<String> &conf, ErrorHandler *errh)
{
    /* conf.size() must be exactly noutputs or noutputs - 1 */
    if (conf.size() != noutputs() &&
	conf.size() + 1 != noutputs()) {
	click_chatter("conf.size() = %d vs noutputs() = %d", conf.size(), noutputs());
	return errh->error("Configuration error");
    }

    if (conf.size() > 1) {
	if (create_program(conf, errh) < 1) {
	    return errh->error("Unable to parse a program");
	}
    }

    return 0;
}

void PSClassify::push(int, Packet *p) 
{
    int size = sips->size();
    int res;
    int i = 0;

#ifdef DEBUG_PSCODE
    click_chatter("Debug");
    Vector<SimpleProg *>::iterator it;
    for(it=sips->begin();it < sips->end(); it++, i++) {
	click_chatter("Program %d:", i);
	(*it)->debug();
    }
#endif

    for(i=0; i < size; i++) {
	SimpleProg *sp = (*sips)[i];
	res = sp->run(p);
	if (res != -1) {
	    output(res).push(p);
	    return;
	}
    }
    output(noutputs()-1).push(p);
    return;
}


SimpleProg::SimpleProg()
{
    ins = new Vector<SimpleIns*>();
}

SimpleProg::~SimpleProg()
{
    delete ins;
}

void SimpleProg::debug()
{

    Vector<SimpleIns*>::iterator it;
    for(it=ins->begin();it<ins->end();it++) {
	click_chatter("%d ", (*it)->getCode());
    }
    click_chatter("--------------------------");

}

void SimpleProg::add_cmd(int code, int output) 
{
    SimpleIns *si = new SimpleIns(code, output);
    ins->push_back(si);
}

bool SimpleProg::terminate_cmd(int output)
{
    SimpleIns *si = ins->back();
    if (si) {
	if (si->getCode() < 1 || si->getCode() > 4) {
	    click_chatter("Error! Program must end with one of (PUB, SUB, META, DATA)");
	    return false;
	}
    }
    add_cmd(128, output);
    return true;
}

int SimpleProg::run(Packet *p)
{
    struct regs mem = { 0, 0, false, false };
    unsigned int ilen = ins->size();

    if (ilen < 1)
	return -1;

    for(mem.pc=0; mem.pc<ilen; mem.pc++) {
	int res = (*ins)[mem.pc]->execute(mem, p);
	click_chatter("State: PC: %d, Res: %d, Reg: %d, Negate=%d, skip=%d", mem.pc, res, mem.reg, (int)mem.negate, (int)mem.skip);
	if (res < 0)
	    return -1;
	if (res == 0)
	    return mem.reg;
	if (mem.skip)
	    skip_operand(mem);
    }
	    
    return mem.reg;
}

void SimpleProg::skip_operand(struct SimpleProg::regs &mem) {
    /* available? */
    int avail = ins->size() - mem.pc - 1;
    if (avail > 0) {
	SimpleIns *si = (*ins)[mem.pc+1];
	/* negation */
	if (si->getCode() == 7)
	    mem.pc += 2;
	else
	    mem.pc += 1;
    }
    mem.skip = false;
}

SimpleIns::SimpleIns(int code, int output) 
{
    mycode = code;
    myout = output;
}

SimpleIns::~SimpleIns() {}

int SimpleIns::getCode() { return mycode; }

    

int SimpleIns::execute(struct SimpleProg::regs &mem, Packet *p)
{
    unsigned char type = p->anno_u8(PS_ANNO_PKTTYPE);

#ifdef DEBUG_PSCODE
    click_chatter("Executing command: %x, type: %x", mycode, (int)type);
#endif
    switch(mycode) {
    case 1: /* PUB */
	if (((type & 0x81) == 0x80) ^ mem.negate)
	    mem.reg = myout;
	else
	    mem.reg = -1;
	mem.negate = false;
	return 1;
    case 2: /* SUB */
	if (((type & 0x81) == 0x81) ^ mem.negate)
	    mem.reg = myout;
	else
	    mem.reg = -1;
	mem.negate = false;
	return 1;
    case 3: /* META */
	if (((type & 0x82) == 0x80) ^ mem.negate)
	    mem.reg = myout;
	else
	    mem.reg = -1;
	mem.negate = false;
	return 1;
    case 4: /* DATA */
	if (((type & 0x82) == 0x82) ^ mem.negate)
	    mem.reg = myout;
	else
	    mem.reg = -1;
	mem.negate = false;
	return 1;
    case 5: /* && */
	if (mem.reg == -1)
	    // skip right operand
	    mem.skip = true;
	return 1;
    case 6: /* || */
	if (mem.reg != -1) 
	    // skip the right operand */
	    mem.skip = true;
	return 1;
    case 7: /* ! */
	mem.negate = true;
	return 1;
    case 128: /* END */
	/* evaluate expression */
	return 0;
    }
    return -1;
}


CLICK_ENDDECLS
EXPORT_ELEMENT(PSClassify)
